# type: ignore
import uvicorn
from fastapi import Depends, FastAPI, Request
from fastapi.security import HTTPBearer
from starlette.responses import PlainTextResponse

from examples import logsetup
from flaat.fastapi import Flaat
from flaat.requirements import HasAARCEntitlement, HasGroup, OneOf, ValidLogin
from flaat.user_infos import UserInfos

logger = logsetup.setup_logging()


##########
## Basic config
# FastAPI
app = FastAPI()
# default_response_class=HTMLResponse
security = HTTPBearer()

# FLAAT
flaat = Flaat()
flaat.set_cache_lifetime(120)  # seconds; default is 300
flaat.set_trusted_OP_list(
    [
        "https://b2access.eudat.eu/oauth2/",
        "https://b2access-integration.fz-juelich.de/oauth2",
        "https://unity.helmholtz-data-federation.de/oauth2/",
        "https://login.helmholtz-data-federation.de/oauth2/",
        "https://login-dev.helmholtz.de/oauth2/",
        "https://login.helmholtz.de/oauth2/",
        "https://unity.eudat-aai.fz-juelich.de/oauth2/",
        "https://services.humanbrainproject.eu/oidc/",
        "https://accounts.google.com/",
        "https://login.elixir-czech.org/oidc/",
        "https://iam-test.indigo-datacloud.eu/",
        "https://iam.deep-hybrid-datacloud.eu/",
        "https://iam.extreme-datacloud.eu/",
        "https://aai.egi.eu/oidc/",
        "https://aai-demo.egi.eu/oidc",
        "https://aai-dev.egi.eu/oidc",
        "https://oidc.scc.kit.edu/auth/realms/kit/",
        "https://proxy.demo.eduteams.org",
        "https://wlcg.cloud.cnaf.infn.it/",
    ]
)
# flaat.set_trusted_OP_file('/etc/oidc-agent/issuer.config')
# flaat.set_OP_hint("helmholtz")
# flaat.set_OP_hint("google")

# verbosity:
#     0: No output
#     1: Errors
#     2: More info, including token info
#     3: Max
# flaat.set_verbosity(0)
# flaat.set_verify_tls(True)


# # Required for using token introspection endpoint:
# flaat.set_client_id('')
# flaat.set_client_secret('')

# flaat.raise_error_on_return = False


def my_failure_callback(message=""):
    return f"User define failure callback.\nError Message: {message}"


@app.get("/")
async def root(request: Request):
    text = """This is an example for useing flaat with FastAPI. These endpoints are available:
    /user_info          General info about the access_token (if provided)
    /user               User built using method
    /valid_user         Requires a valid user
    /valid_user_2       Requires a valid user, uses a custom callback on error
    /group_test_kit     Requires user to have two "eduperson_scoped_affiliation" of
                            ['admins@kit.edu', 'employee@kit.edu', 'member@kit.edu'],
    /group_test_iam     Requires user to be in the group "KIT-Cloud" transported in "groups"
    /group_test_hdf     Requires user to be in all groups found in "eduperson_entitlement"
                            ['urn:geant:h-df.de:group:aai-admin', 'urn:geant:h-df.de:group:myExampleColab#unity.helmholtz-data-federation.de']

    /group_test_hdf2     Requires user to be in all groups found in "eduperson_entitlement"
                            ['urn:geant:h-df.de:group:myExampleColab#unity.helmholtz-data-federation.de'],
    /group_test_hdf3     Requires user to be in all groups found in "eduperson_entitlement"
                            ['urn:geant:h-df.de:group:aai-admin'],
    /group_test_hack    A hack to use any other field for authorisation
    /group_test_wlcg    Requires user to be in the '/wlcg' group
        """
    return PlainTextResponse(text)


@app.get("/info")
@flaat.inject_user_infos
async def user_info(request: Request, user_infos=None):
    if user_infos is not None:
        return {"data": user_infos.__dict__}

    return {"message": "no userinfo"}


def get_user(user_infos: UserInfos):
    return {"infos": user_infos}


@app.get("/info")
@flaat.inject_user_infos
async def user(request: Request, user_infos=None):
    if user_infos is not None:
        return {"data": user_infos.__dict__}

    return {"message": "no userinfo"}


@app.get("/valid_user", dependencies=[Depends(security)])
@flaat.requires(ValidLogin())
def valid_user(request: Request):
    return {"message": "This worked: there was a valid login"}


@app.get("/valid_user_2", dependencies=[Depends(security)])
@flaat.requires(ValidLogin(), on_failure=my_failure_callback)
async def valid_user_own_callback(request: Request):
    return {"message": "This worked: there was a valid login"}


@app.get("/group_test_kit", dependencies=[Depends(security)])
@flaat.requires(
    HasGroup(
        required=["admins@kit.edu", "employee@kit.edu", "member@kit.edu"],
        claim="eduperson_scoped_affiliation",
        match=2,
    ),
    on_failure=my_failure_callback,
)
async def demo_groups_kit(request: Request):
    return {"message": "This worked: user is member of the requested group"}


@app.get("/group_test_iam", dependencies=[Depends(security)])
@flaat.requires(HasGroup("KIT-Cloud", "groups"))
async def demo_groups_iam(request: Request):
    return {"message": "This worked: user is member of the requested group"}


@app.get("/group_test_hdf", dependencies=[Depends(security)])
@flaat.requires(
    HasAARCEntitlement(
        required=[
            "urn:geant:h-df.de:group:m-team:feudal-developers",
            "urn:geant:h-df.de:group:MyExampleColab#unity.helmholtz.de",
        ],
        claim="eduperson_entitlement",
        match="all",
    )
)
async def demo_groups_hdf(request: Request):
    return {"message": "This worked: user has the required entitlement(s)"}


@app.get("/group_test_hdf2", dependencies=[Depends(security)])
@flaat.requires(
    HasAARCEntitlement(
        "urn:geant:h-df.de:group:MyExampleColab", "eduperson_entitlement"
    )
)
async def demo_groups_hdf2(request: Request):
    return {"message": "This worked: user has the required entitlement(s)"}


@app.get("/group_test_hdf3", dependencies=[Depends(security)])
@flaat.requires(
    HasAARCEntitlement(
        [
            "urn:geant:h-df.de:group:MyExampleColab",
            "urn:geant:h-df.de:group:m-team:feudal-developers",
        ],
        "eduperson_entitlement",
    )
)
async def demo_groups_hdf3(request: Request):
    return {"message": "This worked: user has the required entitlement(s)"}


@app.get("/group_test_hack", dependencies=[Depends(security)])
@flaat.requires(
    OneOf(
        HasGroup("Hardt", "family_name"),
        HasGroup("/wlcg", "wlcg.groups"),
    )
)
@app.get("/group_test_wlcg", dependencies=[Depends(security)])
async def demo_groups_wlcg(request: Request):
    return {"message": "This worked: user has the required Group Membership"}


@app.get("/role_test_egi", dependencies=[Depends(security)])
@flaat.requires(
    HasAARCEntitlement(
        "urn:mace:egi.eu:group:mteam.data.kit.edu:role=member",
        "eduperson_entitlement",
    )
)
async def demo_role_egi(request: Request):
    return {"message": "This worked: user is member of the requested group and role"}


##########
# Main

if __name__ == "__main__":
    uvicorn.run(
        "examples.example_fastapi:app", host="0.0.0.0", port=8082, log_level="info"
    )

# start with:
#   python3 example-fastapi.py
# or
#   uvicorn example-fastapi:app --host 0.0.0.0 --port 8082



# This code is distributed under the MIT License

import base64
from dataclasses import dataclass
import logging
from typing import Optional

import jwt
from jwt import PyJWKClient

from flaat.exceptions import FlaatUnauthenticated
from flaat.issuers import IssuerConfig

logger = logging.getLogger(__name__)

# Expand this list in a sensible way
PERMITTED_SIGNATURE_ALGORITHMS = [
    "RS256",
    "RS384",
    "RS512",
]


def _base64_url_encode(data):
    """Decode base64 encode data"""
    if not isinstance(data, bytes):
        data = data.encode("utf-8")
    encode = base64.urlsafe_b64encode(data)
    return encode.decode("utf-8").rstrip("=")


@dataclass
class AccessTokenInfo:
    """Infos from a JWT access token"""

    header: dict
    """ The JWTs JOSE header """

    body: dict
    """ The JWTs data payload """

    signature: str
    """ The JWTs JWS signature """

    verification: Optional[dict]
    """ Infos about the verification of the JWT.
    If set to `None`, then the JWT data is unverified."""

    def __init__(self, complete_decode, verification=Optional[dict]):
        self.header = complete_decode.get("header", {})
        self.body = complete_decode.get("payload", {})
        self.signature = _base64_url_encode(complete_decode.get("signature", b""))
        self.verification = verification

    @property
    def issuer(self) -> str:
        return self.body.get("iss", "")


def _get_signing_key_from_jwt(jwks_uri, access_token) -> str:
    jwk_client = PyJWKClient(jwks_uri)

    try:
        return jwk_client.get_signing_key_from_jwt(access_token).key
    except jwt.exceptions.PyJWKClientError:
        # keys don't have 'kid' field, retrieve key by type, given the 'alg' in token header
        unverified = jwt.api_jwt.decode_complete(
            access_token, options={"verify_signature": False}
        )
        alg = unverified["header"][
            "alg"
        ]  # MUST be present, possible values defined at https://datatracker.ietf.org/doc/html/rfc7518#section-3.1

        # algorithm is none, then signing key is None; signature must be empty octet string
        if alg == "none":
            return ""
        # infer key type from algorithm
        key_type = ""
        if alg.startswith("RS") or alg.startswith("PS"):
            key_type = "RSA"
        if alg.startswith("HS"):
            key_type = "oct"
        if alg.startswith("ES"):
            key_type = "EC"
        if alg.startswith("Ed"):
            key_type = "OKP"
        # get key from JWKS endpoint by key type
        jwk_set = jwk_client.get_jwk_set()
        signing_keys = [
            jwk_set_key
            for jwk_set_key in jwk_set.keys
            if jwk_set_key.key_type == key_type
        ]
        if len(signing_keys) == 0:
            raise FlaatUnauthenticated(
                "Could not verify JWT: The JWKS endpoint did not contain any signing key "
                f"of type {key_type}."
            )
        if len(signing_keys) > 1:
            raise FlaatUnauthenticated(
                f"Could not verify JWT: The JWKS endpoint has too many keys of type {key_type} "
                "and no 'kid' specified."
            )
        return signing_keys[0].key


def get_access_token_info(access_token, verify=True) -> Optional[AccessTokenInfo]:
    unverified = {}
    try:
        unverified = jwt.api_jwt.decode_complete(
            access_token,
            options={
                "verify_signature": False,
            },
        )
        unverified_body = unverified.get("payload", {})
    except jwt.DecodeError:
        return None

    if not verify:
        return AccessTokenInfo(unverified, verification=None)

    issuer = IssuerConfig.get_from_string(unverified_body.get("iss", ""))
    if issuer is None:
        raise FlaatUnauthenticated("Could not verify JWT: No 'iss' claim in body")

    jwks_uri = issuer.issuer_config.get("jwks_uri", "")
    if jwks_uri == "":
        raise FlaatUnauthenticated(
            "Could not verify JWT: Issuer config has no jwks_uri"
        )

    signing_key = _get_signing_key_from_jwt(jwks_uri, access_token)

    try:
        complete_decode = jwt.api_jwt.decode_complete(
            access_token,
            signing_key,
            algorithms=PERMITTED_SIGNATURE_ALGORITHMS,
            options={"verify_aud": False},
        )
    except jwt.exceptions.PyJWTError as e:
        raise FlaatUnauthenticated(f"Could not verify JWT: {e}") from e

    return AccessTokenInfo(
        complete_decode,
        verification={"algorithm": complete_decode.get("header", {}).get("alg", "")},
    )



import pytest

from flaat.access_tokens import get_access_token_info
from flaat.test_env import FLAAT_AT, NON_JWT_FLAAT_AT


class TestTokens:
    def test_get_accesstoken_info_non_jwt(self):
        access_token = NON_JWT_FLAAT_AT
        if access_token == "":
            pytest.skip("No non JWT access token")

        access_token_info = get_access_token_info(access_token)
        assert access_token_info is None

    def test_get_accesstoken_info_jwt(self):
        access_token = FLAAT_AT
        if access_token in ["", "mock_jwt_at"]:
            pytest.skip("No JWT access token")
        access_token_info = get_access_token_info(access_token)
        assert access_token_info is not None
        assert access_token_info.verification is not None



import json
import logging

from flask import request
from flask.wrappers import Response

from flaat import BaseFlaat

logger = logging.getLogger(__name__)


class Flaat(BaseFlaat):
    def _get_request(self, *_, **__):
        return request

    def _get_header_from_request(self, _, name) -> str:
        # using flask global "request" here, not an argument
        return request.headers.get(name, "")

    def _make_response(self, data, status_code: int):
        return Response(
            response=json.dumps(data),
            status=status_code,
            mimetype="application/json",
        )



from flask import Flask, request
from werkzeug import Response

from examples import logsetup
from flaat.flask import Flaat
from flaat.requirements import HasAARCEntitlement, HasGroup, ValidLogin

logger = logsetup.setup_logging()


##########
## Basic config
# FLASK
app = Flask(__name__)

# FLAAT
flaat = Flaat()
flaat.set_cache_lifetime(120)  # seconds; default is 300
flaat.set_trusted_OP_list(
    [
        "https://aai-demo.egi.eu/oidc",
        "https://aai-dev.egi.eu/oidc",
        "https://aai.egi.eu/oidc/",
        "https://accounts.google.com/",
        "https://b2access-integration.fz-juelich.de/oauth2",
        "https://b2access.eudat.eu/oauth2/",
        "https://iam-test.indigo-datacloud.eu/",
        "https://iam.deep-hybrid-datacloud.eu/",
        "https://iam.extreme-datacloud.eu/",
        "https://login-dev.helmholtz.de/oauth2/",
        "https://login.elixir-czech.org/oidc/",
        "https://login.helmholtz-data-federation.de/oauth2/",
        "https://login.helmholtz.de/oauth2/",
        "https://oidc.scc.kit.edu/auth/realms/kit/",
        "https://orcid.org/",
        "https://proxy.demo.eduteams.org",
        "https://services.humanbrainproject.eu/oidc/",
        "https://unity.eudat-aai.fz-juelich.de/oauth2/",
        "https://unity.helmholtz-data-federation.de/oauth2/",
        "https://wlcg.cloud.cnaf.infn.it/",
    ]
)
# flaat.set_trusted_OP_file('/etc/oidc-agent/issuer.config')
# flaat.set_OP_hint("helmholtz")
# flaat.set_OP_hint("google")
flaat.set_timeout(3)

# verbosity:
#     0: No output
#     1: Errors
#     2: More info, including token info
#     3: Max
# flaat.set_verbosity(0)
# flaat.set_verify_tls(True)


# # Required for using token introspection endpoint:
# flaat.set_client_id('')
# flaat.set_client_secret('')


def my_failure_callback(message=""):
    return f"User define failure callback.\nError Message: {message}"


@app.route("/")
def root():
    text = """This is an example for useing flaat with AIO. These endpoints are available:
    /info               General info about the access_token (if provided)
    /valid_user         Requires a valid user
    /valid_user_2       Requires a valid user, uses a custom callback on error
    /group_test_kit     Requires user to have two "eduperson_scoped_affiliation" of
                            ['admins@kit.edu', 'employee@kit.edu', 'member@kit.edu'],
    /group_test_iam     Requires user to be in the group "KIT-Cloud" transported in "groups"
    /group_test_hdf     Requires user to be in all groups found in "eduperson_entitlement"
                            ['urn:geant:h-df.de:group:aai-admin', 'urn:geant:h-df.de:group:myExampleColab#unity.helmholtz-data-federation.de']

    /group_test_hdf2     Requires user to be in all groups found in "eduperson_entitlement"
                            ['urn:geant:h-df.de:group:myExampleColab#unity.helmholtz-data-federation.de'],
    /group_test_hdf3     Requires user to be in all groups found in "eduperson_entitlement"
                            ['urn:geant:h-df.de:group:aai-admin'],
    /group_test_hack    A hack to use any other field for authorisation
    /group_test_wlcg    Requires user to be in the '/wlcg' group
        """
    return Response(text, mimetype="text/plain")


@app.route("/info")
@flaat.inject_user_infos
def info(user_infos=None):
    if user_infos is not None:
        return user_infos.toJSON()
    return "No user infos"


@app.route("/valid_user/<int:id>", methods=["POST", "GET"])
@flaat.requires(ValidLogin())
def valid_user_id(id):
    retval = ""
    if request.method == "POST":
        retval += "POST\n"
    if request.method == "GET":
        retval += "GET\n"
    retval += f"This worked: there was a valid login, and an id: {id}\n"
    return retval


@app.route("/valid_user")
@flaat.requires(ValidLogin())
def valid_user():
    return "This worked: there was a valid login\n"


@app.route("/valid_user_2")
@flaat.requires(ValidLogin(), on_failure=my_failure_callback)
def valid_user_own_callback():
    return "This worked: there was a valid login"


@app.route("/group_test_kit")
@flaat.requires(
    HasGroup(
        required=["admins@kit.edu", "employee@kit.edu", "member@kit.edu"],
        claim="eduperson_scoped_affiliation",
        match=2,
    ),
    on_failure=my_failure_callback,
)
def demo_groups_kit():
    return "This worked: user is member of the requested group"


@app.route("/group_test_iam")
@flaat.requires(HasGroup("KIT-Cloud", "groups"))
def demo_groups_iam():
    return "This worked: user is member of the requested group"


@app.route("/group_test_hdf")
@flaat.requires(
    HasAARCEntitlement(
        required=[
            "urn:geant:h-df.de:group:m-team:feudal-developers",
            "urn:geant:h-df.de:group:MyExampleColab#unity.helmholtz.de",
        ],
        claim="eduperson_entitlement",
        match="all",
    )
)
def demo_groups_hdf():
    return "This worked: user is member of the requested group"


@app.route("/group_test_hdf2")
@flaat.requires(
    HasAARCEntitlement(
        "urn:geant:h-df.de:group:MyExampleColab", "eduperson_entitlement"
    )
)
def demo_groups_hdf2():
    return "This worked: user is member of the requested group"


@app.route("/group_test_hdf3")
@flaat.requires(
    HasAARCEntitlement(
        [
            "urn:geant:h-df.de:group:MyExampleColab",
            "urn:geant:h-df.de:group:m-team:feudal-developers",
        ],
        "eduperson_entitlement",
    )
)
def demo_groups_hdf3():
    return "This worked: user is member of the requested group"


##########
# Main
if __name__ == "__main__":
    # app.run(host="127.0.0.1", port=8081, debug=True)
    app.run(host="0.0.0.0", port=8081, debug=True)



# pylint: disable=redefined-outer-name
import pytest
from flask.app import Flask
from werkzeug import Response

from flaat.flask import Flaat
from flaat.test_env import FLAAT_TRUSTED_OPS_LIST, User, get_status_kwargs_list

flaat = Flaat()
flaat.set_trusted_OP_list(FLAAT_TRUSTED_OPS_LIST)

DECORATORS = User(flaat).get_named_decorators()


def view_func(test_inject=None):
    _ = test_inject
    return Response(response="Success")


@pytest.fixture
def app():
    """flask web app for testing"""
    app = Flask(__name__)
    for decorator in DECORATORS:

        decorated_view_func = decorator.decorator(view_func)
        # rename to decorator name, as flask does not allow duplicate view_func names
        decorated_view_func.__name__ = f"{decorator.name}-view_func"

        app.route(f"/{decorator.name}")(decorated_view_func)

    return app


@pytest.fixture
def client(app: Flask):
    return app.test_client()


@pytest.mark.parametrize("status,kwargs", get_status_kwargs_list())
@pytest.mark.parametrize("decorator", DECORATORS)
def test_decorator(client, decorator, status, kwargs):
    resp = client.get(f"/{decorator.name}", **kwargs)
    expected = decorator.get_expected_status_code(status)
    assert resp.status_code == expected



import logging
import sys


def setup_logging():
    """setup logging"""
    formatter = logging.Formatter("%(levelname)s - %(message)s")

    handler = logging.StreamHandler(sys.stderr)
    handler.setFormatter(formatter)

    rootLogger = logging.getLogger()
    rootLogger.addHandler(handler)
    rootLogger.setLevel(logging.DEBUG)

    # turn off other logging:
    for other in ["werkzeug", "urllib3"]:
        other_log = logging.getLogger(other)
        other_log.setLevel(logging.CRITICAL)
